@@ -26,7 +26,7 @@ from ..common.objects import (
 )
 from ..common.auth import ClusterAuth
 from ..common.utils import (
-    namespace_default,
+    get_current_namespace,
     escape,
 )
 from ..common.networking import (
@@ -97,7 +97,7 @@ class Pod(ProcessInterface):
             except ApiException as e:
                 if e.reason == "Not Found":
                     logger.debug(
-                        "Pod %s in namespace %s has been deleated already.",
+                        "Pod %s in namespace %s has been deleted already.",
                         name,
                         namespace,
                     )
@@ -582,10 +582,10 @@ class KubeCluster(SpecCluster):
         await ClusterAuth.load_first(self.auth)
 
         self.core_api = kubernetes.client.CoreV1Api()
-        self.policy_api = kubernetes.client.PolicyV1beta1Api()
+        self.policy_api = kubernetes.client.PolicyV1Api()
 
         if self.namespace is None:
-            self.namespace = namespace_default()
+            self.namespace = get_current_namespace()
 
         environ = {k: v for k, v in os.environ.items() if k not in ["user", "uuid"]}
         self._generate_name = self._generate_name.format(
@@ -644,12 +644,15 @@ class KubeCluster(SpecCluster):
 
         await super()._start()
 
-        dashboard_address = await get_scheduler_address(
-            self.scheduler.service.metadata.name,
-            self.namespace,
-            port_name="http-dashboard",
-        )
-        self.forwarded_dashboard_port = dashboard_address.split(":")[-1]
+        if self._deploy_mode == "local":
+            self.forwarded_dashboard_port = self.scheduler.services["dashboard"].port
+        else:
+            dashboard_address = await get_scheduler_address(
+                self.scheduler.service.metadata.name,
+                self.namespace,
+                port_name="http-dashboard",
+            )
+            self.forwarded_dashboard_port = dashboard_address.split(":")[-1]
 
     @classmethod
     def from_dict(cls, pod_spec, **kwargs):
