@@ -4,12 +4,18 @@ import asyncio
 import atexit
 from contextlib import suppress
 from enum import Enum
+import getpass
+import os
 import time
 from typing import ClassVar
+import warnings
 import weakref
+import uuid
 
 import kubernetes_asyncio as kubernetes
+import yaml
 
+import dask.config
 from distributed.core import Status, rpc
 from distributed.deploy import Cluster
 from distributed.utils import (
@@ -20,9 +26,7 @@ from distributed.utils import (
 )
 
 from dask_kubernetes.common.auth import ClusterAuth
-from dask_kubernetes.common.utils import namespace_default
 from dask_kubernetes.operator import (
-    build_cluster_spec,
     wait_for_service,
 )
 
@@ -31,6 +35,7 @@ from dask_kubernetes.common.networking import (
     wait_for_scheduler,
     wait_for_scheduler_comm,
 )
+from dask_kubernetes.common.utils import get_current_namespace
 
 
 class CreateMode(Enum):
@@ -85,6 +90,17 @@ class KubeCluster(Cluster):
     shutdown_on_close: bool (optional)
         Whether or not to delete the cluster resource when this object is closed.
         Defaults to ``True`` when creating a cluster and ``False`` when connecting to an existing one.
+    resource_timeout: int (optional)
+        Time in seconds to wait for the controller to take action before giving up.
+        If the ``DaskCluster`` resource that gets created isn't moved into a known ``status.phase`` by the controller
+        then it is likely the controller isn't running or is malfunctioning and we time out and clean up with a
+        useful error.
+        Defaults to ``60`` seconds.
+    scheduler_service_type: str (optional)
+        Kubernetes service type to use for the scheduler. Defaults to ``ClusterIP``.
+    custom_cluster_spec: str | dict (optional)
+        Path to a YAML manifest or a dictionary representation of a ``DaskCluster`` resource object which will be
+        used to create the cluster instead of generating one from the other keyword arguments.
     **kwargs: dict
         Additional keyword arguments to pass to LocalCluster
 
@@ -125,31 +141,75 @@ class KubeCluster(Cluster):
 
     def __init__(
         self,
-        name,
+        name=None,
         namespace=None,
-        image="ghcr.io/dask/dask:latest",
-        n_workers=3,
-        resources={},
-        env=[],
-        worker_command="dask-worker",
+        image=None,
+        n_workers=None,
+        resources=None,
+        env=None,
+        worker_command=None,
         auth=ClusterAuth.DEFAULT,
         port_forward_cluster_ip=None,
-        create_mode=CreateMode.CREATE_OR_CONNECT,
+        create_mode=None,
         shutdown_on_close=None,
+        resource_timeout=None,
+        scheduler_service_type=None,
+        custom_cluster_spec=None,
         **kwargs,
     ):
-        self.namespace = namespace or namespace_default()
-        self.image = image
-        self.n_workers = n_workers
-        self.resources = resources
-        self.env = env
-        self.worker_command = worker_command
+        name = dask.config.get("kubernetes.name", override_with=name)
+        self.namespace = (
+            dask.config.get("kubernetes.namespace", override_with=namespace)
+            or get_current_namespace()
+        )
+        self.image = dask.config.get("kubernetes.image", override_with=image)
+        self.n_workers = dask.config.get(
+            "kubernetes.count.start", override_with=n_workers
+        )
+        if dask.config.get("kubernetes.count.max"):
+            warnings.warn(
+                "Setting a maximum number of workers is no longer supported. "
+                "Please use Kubernetes Resource Quotas instead."
+            )
+        self.resources = dask.config.get(
+            "kubernetes.resources", override_with=resources
+        )
+        self.env = dask.config.get("kubernetes.env", override_with=env)
+        self.worker_command = dask.config.get(
+            "kubernetes.worker-command", override_with=worker_command
+        )
+        self.auth = auth
+        self.port_forward_cluster_ip = dask.config.get(
+            "kubernetes.port-forward-cluster-ip", override_with=port_forward_cluster_ip
+        )
+        self.create_mode = dask.config.get(
+            "kubernetes.create-mode", override_with=create_mode
+        )
+        self.shutdown_on_close = dask.config.get(
+            "kubernetes.shutdown-on-close", override_with=shutdown_on_close
+        )
+        self._resource_timeout = dask.config.get(
+            "kubernetes.resource-timeout", override_with=resource_timeout
+        )
+        self._custom_cluster_spec = dask.config.get(
+            "kubernetes.custom-cluster-spec", override_with=custom_cluster_spec
+        )
+        self.scheduler_service_type = dask.config.get(
+            "kubernetes.scheduler-service-type", override_with=scheduler_service_type
+        )
+
+        if self._custom_cluster_spec:
+            if isinstance(self._custom_cluster_spec, str):
+                with open(self._custom_cluster_spec) as f:
+                    self._custom_cluster_spec = yaml.safe_load(f.read())
+            name = self._custom_cluster_spec["metadata"]["name"]
+
         if isinstance(self.worker_command, str):
             self.worker_command = self.worker_command.split(" ")
-        self.auth = auth
-        self.port_forward_cluster_ip = port_forward_cluster_ip
-        self.create_mode = create_mode
-        self.shutdown_on_close = shutdown_on_close
+
+        name = name.format(
+            user=getpass.getuser(), uuid=str(uuid.uuid4())[:10], **os.environ
+        )
 
         self._instances.add(self)
 
@@ -158,10 +218,6 @@ class KubeCluster(Cluster):
             self._loop_runner.start()
             self.sync(self._start)
 
-    @property
-    def cluster_name(self):
-        return f"{self.name}-cluster"
-
     @property
     def dashboard_link(self):
         host = self.scheduler_address.split("://")[1].split("/")[0].split(":")[0]
@@ -173,13 +229,13 @@ class KubeCluster(Cluster):
 
         if cluster_exists and self.create_mode == CreateMode.CREATE_ONLY:
             raise ValueError(
-                f"Cluster {self.cluster_name} already exists and create mode is '{CreateMode.CREATE_ONLY}'"
+                f"Cluster {self.name} already exists and create mode is '{CreateMode.CREATE_ONLY}'"
             )
         elif cluster_exists:
             await self._connect_cluster()
         elif not cluster_exists and self.create_mode == CreateMode.CONNECT_ONLY:
             raise ValueError(
-                f"Cluster {self.cluster_name} doesn't and create mode is '{CreateMode.CONNECT_ONLY}'"
+                f"Cluster {self.name} doesn't and create mode is '{CreateMode.CONNECT_ONLY}'"
             )
         else:
             await self._create_cluster()
@@ -193,12 +249,18 @@ class KubeCluster(Cluster):
             core_api = kubernetes.client.CoreV1Api(api_client)
             custom_objects_api = kubernetes.client.CustomObjectsApi(api_client)
 
-            service_name = f"{self.name}-cluster-service"
-            cluster_name = f"{self.name}-cluster"
-            worker_spec = self._build_worker_spec(service_name)
-            scheduler_spec = self._build_scheduler_spec(cluster_name)
-
-            data = build_cluster_spec(cluster_name, worker_spec, scheduler_spec)
+            if not self._custom_cluster_spec:
+                data = make_cluster_spec(
+                    name=self.name,
+                    env=self.env,
+                    resources=self.resources,
+                    worker_command=self.worker_command,
+                    n_workers=self.n_workers,
+                    image=self.image,
+                    scheduler_service_type=self.scheduler_service_type,
+                )
+            else:
+                data = self._custom_cluster_spec
             try:
                 await custom_objects_api.create_namespaced_custom_object(
                     group="kubernetes.dask.org",
@@ -213,13 +275,19 @@ class KubeCluster(Cluster):
                     "Are the Dask Custom Resource Definitions installed? "
                     "https://kubernetes.dask.org/en/latest/operator.html#installing-the-operator"
                 ) from e
-            await wait_for_scheduler(cluster_name, self.namespace)
-            await wait_for_service(core_api, f"{cluster_name}-service", self.namespace)
+
+            try:
+                await self._wait_for_controller()
+            except TimeoutError as e:
+                await self._close()
+                raise e
+            await wait_for_scheduler(self.name, self.namespace)
+            await wait_for_service(core_api, f"{self.name}-scheduler", self.namespace)
             scheduler_address = await self._get_scheduler_address()
             await wait_for_scheduler_comm(scheduler_address)
             self.scheduler_comm = rpc(scheduler_address)
             dashboard_address = await get_scheduler_address(
-                f"{self.name}-cluster-service",
+                f"{self.name}-scheduler",
                 self.namespace,
                 port_name="http-dashboard",
             )
@@ -242,8 +310,8 @@ class KubeCluster(Cluster):
                 self.env = container_spec["env"]
             else:
                 self.env = {}
-            service_name = f'{cluster_spec["metadata"]["name"]}-service'
-            await wait_for_scheduler(self.cluster_name, self.namespace)
+            service_name = f"{cluster_spec['metadata']['name']}-scheduler"
+            await wait_for_scheduler(self.name, self.namespace)
             await wait_for_service(core_api, service_name, self.namespace)
             scheduler_address = await self._get_scheduler_address()
             await wait_for_scheduler_comm(scheduler_address)
@@ -264,26 +332,47 @@ class KubeCluster(Cluster):
                     version="v1",
                     plural="daskclusters",
                     namespace=self.namespace,
-                    name=self.cluster_name,
+                    name=self.name,
                 )
             except kubernetes.client.exceptions.ApiException as e:
                 return None
 
     async def _get_scheduler_address(self):
-        service_name = f"{self.name}-cluster-service"
-        address = await get_scheduler_address(service_name, self.namespace)
+        address = await get_scheduler_address(f"{self.name}-scheduler", self.namespace)
         return address
 
+    async def _wait_for_controller(self):
+        """Wait for the operator to set the status.phase."""
+        async with kubernetes.client.api_client.ApiClient() as api_client:
+            custom_objects_api = kubernetes.client.CustomObjectsApi(api_client)
+            watch = kubernetes.watch.Watch()
+            async for event in watch.stream(
+                func=custom_objects_api.list_namespaced_custom_object,
+                group="kubernetes.dask.org",
+                version="v1",
+                plural="daskclusters",
+                namespace=self.namespace,
+                field_selector=f"metadata.name={self.name}",
+                timeout_seconds=self._resource_timeout,
+            ):
+                cluster = event["object"]
+                if "status" in cluster and "phase" in cluster["status"]:
+                    return
+                await asyncio.sleep(0.1)
+        raise TimeoutError(
+            f"Dask Cluster resource not actioned after {self._resource_timeout} seconds, is the Dask Operator running?"
+        )
+
     def get_logs(self):
         """Get logs for Dask scheduler and workers.
 
         Examples
         --------
         >>> cluster.get_logs()
-        {'foo-cluster-scheduler': ...,
-        'foo-cluster-default-worker-group-worker-0269dbfa0cfd4a22bcd9d92ae032f4d2': ...,
-        'foo-cluster-default-worker-group-worker-7c1ccb04cd0e498fb21babaedd00e5d4': ...,
-        'foo-cluster-default-worker-group-worker-d65bee23bdae423b8d40c5da7a1065b6': ...}
+        {'foo': ...,
+        'foo-default-worker-0269dbfa0cfd4a22bcd9d92ae032f4d2': ...,
+        'foo-default-worker-7c1ccb04cd0e498fb21babaedd00e5d4': ...,
+        'foo-default-worker-d65bee23bdae423b8d40c5da7a1065b6': ...}
         Each log will be a string of all logs for that container. To view
         it is recommeded that you print each log.
         >>> print(cluster.get_logs()["testdask-scheduler-5c8ffb6b7b-sjgrg"])
@@ -303,7 +392,7 @@ class KubeCluster(Cluster):
 
             pods = await core_api.list_namespaced_pod(
                 namespace=self.namespace,
-                label_selector=f"dask.org/cluster-name={self.name}-cluster",
+                label_selector=f"dask.org/cluster-name={self.name}",
             )
 
             for pod in pods.items:
@@ -324,7 +413,16 @@ class KubeCluster(Cluster):
 
         return logs
 
-    def add_worker_group(self, name, n_workers=3, image=None, resources=None, env=None):
+    def add_worker_group(
+        self,
+        name,
+        n_workers=3,
+        image=None,
+        resources=None,
+        worker_command=None,
+        env=None,
+        custom_spec=None,
+    ):
         """Create a dask worker group by name
 
         Parameters
@@ -343,6 +441,9 @@ class KubeCluster(Cluster):
         env: List[dict]
             List of environment variables to pass to worker pod.
             If ommitted will use the cluster default.
+        custom_spec: dict (optional)
+            A dictionary representation of a worker spec which will be used to create the ``DaskWorkerGroup`` instead
+            of generating one from the other keyword arguments.
 
         Examples
         --------
@@ -354,20 +455,38 @@ class KubeCluster(Cluster):
             n_workers=n_workers,
             image=image,
             resources=resources,
+            worker_command=worker_command,
             env=env,
+            custom_spec=custom_spec,
         )
 
     async def _add_worker_group(
-        self, name, n_workers=3, image=None, resources=None, env=None
+        self,
+        name,
+        n_workers=3,
+        image=None,
+        resources=None,
+        worker_command=None,
+        env=None,
+        custom_spec=None,
     ):
-        service_name = f"{self.cluster_name}-service"
-        spec = self._build_worker_spec(service_name)
+        if custom_spec is not None:
+            spec = custom_spec
+        else:
+            spec = make_worker_spec(
+                cluster_name=self.name,
+                env=env or self.env,
+                resources=resources or self.resources,
+                worker_command=worker_command or self.worker_command,
+                n_workers=n_workers or self.n_workers,
+                image=image or self.image,
+            )
         data = {
             "apiVersion": "kubernetes.dask.org/v1",
             "kind": "DaskWorkerGroup",
-            "metadata": {"name": f"{self.name}-cluster-{name}"},
+            "metadata": {"name": f"{self.name}-{name}"},
             "spec": {
-                "cluster": f"{self.name}-cluster",
+                "cluster": f"{self.name}",
                 "worker": spec,
             },
         }
@@ -404,7 +523,7 @@ class KubeCluster(Cluster):
                 version="v1",
                 plural="daskworkergroups",
                 namespace=self.namespace,
-                name=f"{self.name}-cluster-{name}",
+                name=f"{self.name}-{name}",
             )
 
     def close(self, timeout=3600):
@@ -421,13 +540,13 @@ class KubeCluster(Cluster):
                     version="v1",
                     plural="daskclusters",
                     namespace=self.namespace,
-                    name=self.cluster_name,
+                    name=self.name,
                 )
             start = time.time()
             while (await self._get_cluster()) is not None:
                 if time.time() > start + timeout:
                     raise TimeoutError(
-                        f"Timed out deleting cluster resource {self.cluster_name}"
+                        f"Timed out deleting cluster resource {self.name}"
                     )
                 await asyncio.sleep(1)
 
@@ -455,111 +574,77 @@ class KubeCluster(Cluster):
             custom_objects_api.api_client.set_default_header(
                 "content-type", "application/merge-patch+json"
             )
+            # Disable adaptivity if enabled
+            with suppress(kubernetes.client.ApiException):
+                await custom_objects_api.delete_namespaced_custom_object(
+                    group="kubernetes.dask.org",
+                    version="v1",
+                    plural="daskautoscalers",
+                    namespace=self.namespace,
+                    name=self.name,
+                )
             await custom_objects_api.patch_namespaced_custom_object_scale(
                 group="kubernetes.dask.org",
                 version="v1",
                 plural="daskworkergroups",
                 namespace=self.namespace,
-                name=f"{self.name}-cluster-{worker_group}-worker-group",
+                name=f"{self.name}-{worker_group}",
                 body={"spec": {"replicas": n}},
             )
 
-    def adapt(self, *args, **kwargs):
-        """Turn on adaptivity"""
-        raise NotImplementedError(
-            "Adaptive mode is not supported yet for this KubeCluster."
-        )
+    def adapt(self, minimum=None, maximum=None):
+        """Turn on adaptivity
 
-    def _build_scheduler_spec(self, cluster_name):
-        # TODO: Take the values provided in the current class constructor
-        # and build a DaskWorker compatible dict
-        if isinstance(self.env, dict):
-            env = [{"name": key, "value": value} for key, value in self.env.items()]
-        else:
-            # If they gave us a list, assume its a list of dicts and already ready to go
-            env = self.env
+        Parameters
+        ----------
+        minimum : int
+            Minimum number of workers
+        minimum : int
+            Maximum number of workers
 
-        return {
-            "spec": {
-                "containers": [
-                    {
-                        "name": "scheduler",
-                        "image": self.image,
-                        "args": ["dask-scheduler", "--host", "0.0.0.0"],
-                        "env": env,
-                        "resources": self.resources,
-                        "ports": [
-                            {
-                                "name": "tcp-comm",
-                                "containerPort": 8786,
-                                "protocol": "TCP",
-                            },
-                            {
-                                "name": "http-dashboard",
-                                "containerPort": 8787,
-                                "protocol": "TCP",
-                            },
-                        ],
-                        "readinessProbe": {
-                            "httpGet": {"port": "http-dashboard", "path": "/health"},
-                            "initialDelaySeconds": 5,
-                            "periodSeconds": 10,
+        Examples
+        --------
+        >>> cluster.adapt()  # Allow scheduler to add/remove workers within k8s cluster resource limits
+        >>> cluster.adapt(minimum=1, maximum=10) # Allow scheduler to add/remove workers within 1-10 range
+        """
+        return self.sync(self._adapt, minimum, maximum)
+
+    async def _adapt(self, minimum=None, maximum=None):
+        async with kubernetes.client.api_client.ApiClient() as api_client:
+            custom_objects_api = kubernetes.client.CustomObjectsApi(api_client)
+            custom_objects_api.api_client.set_default_header(
+                "content-type", "application/merge-patch+json"
+            )
+            try:
+                await custom_objects_api.patch_namespaced_custom_object_scale(
+                    group="kubernetes.dask.org",
+                    version="v1",
+                    plural="daskautoscalers",
+                    namespace=self.namespace,
+                    name=self.name,
+                    body={"spec": {"minimum": minimum, "maximum": maximum}},
+                )
+            except kubernetes.client.ApiException:
+                await custom_objects_api.create_namespaced_custom_object(
+                    group="kubernetes.dask.org",
+                    version="v1",
+                    plural="daskautoscalers",
+                    namespace=self.namespace,
+                    body={
+                        "apiVersion": "kubernetes.dask.org/v1",
+                        "kind": "DaskAutoscaler",
+                        "metadata": {
+                            "name": self.name,
+                            "dask.org/cluster-name": self.name,
+                            "dask.org/component": "autoscaler",
                         },
-                        "livenessProbe": {
-                            "httpGet": {"port": "http-dashboard", "path": "/health"},
-                            "initialDelaySeconds": 15,
-                            "periodSeconds": 20,
+                        "spec": {
+                            "cluster": self.name,
+                            "minimum": minimum,
+                            "maximum": maximum,
                         },
-                    }
-                ]
-            },
-            "service": {
-                "type": "ClusterIP",
-                "selector": {
-                    "dask.org/cluster-name": cluster_name,
-                    "dask.org/component": "scheduler",
-                },
-                "ports": [
-                    {
-                        "name": "tcp-comm",
-                        "protocol": "TCP",
-                        "port": 8786,
-                        "targetPort": "tcp-comm",
                     },
-                    {
-                        "name": "http-dashboard",
-                        "protocol": "TCP",
-                        "port": 8787,
-                        "targetPort": "http-dashboard",
-                    },
-                ],
-            },
-        }
-
-    def _build_worker_spec(self, service_name):
-        if isinstance(self.env, dict):
-            env = [{"name": key, "value": value} for key, value in self.env.items()]
-        else:
-            # If they gave us a list, assume its a list of dicts and already ready to go
-            env = self.env
-
-        args = self.worker_command + ["--name", "$(DASK_WORKER_NAME)"]
-
-        return {
-            "cluster": self.cluster_name,
-            "replicas": self.n_workers,
-            "spec": {
-                "containers": [
-                    {
-                        "name": "worker",
-                        "image": self.image,
-                        "args": args,
-                        "env": env,
-                        "resources": self.resources,
-                    }
-                ]
-            },
-        }
+                )
 
     def __enter__(self):
         return self
@@ -585,6 +670,167 @@ class KubeCluster(Cluster):
         return cls(name=name, create_mode=CreateMode.CONNECT_ONLY, **kwargs)
 
 
+def make_cluster_spec(
+    name,
+    image="ghcr.io/dask/dask:latest",
+    n_workers=None,
+    resources=None,
+    env=None,
+    worker_command="dask-worker",
+    scheduler_service_type="ClusterIP",
+):
+    """Generate a ``DaskCluster`` kubernetes resource.
+
+    Populate a template with some common options to generate a ``DaskCluster`` kubernetes resource.
+
+    Parameters
+    ----------
+    name: str
+        Name of the cluster
+    image: str (optional)
+        Container image to use for the scheduler and workers
+    n_workers: int (optional)
+        Number of workers in the default worker group
+    resources: dict (optional)
+        Resource limits to set on scheduler and workers
+    env: dict (optional)
+        Environment variables to set on scheduler and workers
+    worker_command: str (optional)
+        Worker command to use when starting the workers
+    """
+    return {
+        "apiVersion": "kubernetes.dask.org/v1",
+        "kind": "DaskCluster",
+        "metadata": {"name": name},
+        "spec": {
+            "worker": make_worker_spec(
+                cluster_name=name,
+                env=env,
+                resources=resources,
+                worker_command=worker_command,
+                n_workers=n_workers,
+                image=image,
+            ),
+            "scheduler": make_scheduler_spec(
+                cluster_name=name,
+                env=env,
+                resources=resources,
+                image=image,
+                scheduler_service_type=scheduler_service_type,
+            ),
+        },
+    }
+
+
+def make_worker_spec(
+    cluster_name,
+    image="ghcr.io/dask/dask:latest",
+    n_workers=3,
+    resources=None,
+    env=None,
+    worker_command="dask-worker",
+):
+    if isinstance(env, dict):
+        env = [{"name": key, "value": value} for key, value in env.items()]
+    else:
+        # If they gave us a list, assume its a list of dicts and already ready to go
+        env = env
+
+    if isinstance(worker_command, str):
+        worker_command = worker_command.split(" ")
+
+    args = worker_command + ["--name", "$(DASK_WORKER_NAME)"]
+
+    return {
+        "cluster": cluster_name,
+        "replicas": n_workers,
+        "spec": {
+            "containers": [
+                {
+                    "name": "worker",
+                    "image": image,
+                    "args": args,
+                    "env": env,
+                    "resources": resources,
+                }
+            ]
+        },
+    }
+
+
+def make_scheduler_spec(
+    cluster_name,
+    env=None,
+    resources=None,
+    image="ghcr.io/dask/dask:latest",
+    scheduler_service_type="ClusterIP",
+):
+    # TODO: Take the values provided in the current class constructor
+    # and build a DaskWorker compatible dict
+    if isinstance(env, dict):
+        env = [{"name": key, "value": value} for key, value in env.items()]
+    else:
+        # If they gave us a list, assume its a list of dicts and already ready to go
+        env = env
+
+    return {
+        "spec": {
+            "containers": [
+                {
+                    "name": "scheduler",
+                    "image": image,
+                    "args": ["dask-scheduler", "--host", "0.0.0.0"],
+                    "env": env,
+                    "resources": resources,
+                    "ports": [
+                        {
+                            "name": "tcp-comm",
+                            "containerPort": 8786,
+                            "protocol": "TCP",
+                        },
+                        {
+                            "name": "http-dashboard",
+                            "containerPort": 8787,
+                            "protocol": "TCP",
+                        },
+                    ],
+                    "readinessProbe": {
+                        "httpGet": {"port": "http-dashboard", "path": "/health"},
+                        "initialDelaySeconds": 5,
+                        "periodSeconds": 10,
+                    },
+                    "livenessProbe": {
+                        "httpGet": {"port": "http-dashboard", "path": "/health"},
+                        "initialDelaySeconds": 15,
+                        "periodSeconds": 20,
+                    },
+                }
+            ]
+        },
+        "service": {
+            "type": scheduler_service_type,
+            "selector": {
+                "dask.org/cluster-name": cluster_name,
+                "dask.org/component": "scheduler",
+            },
+            "ports": [
+                {
+                    "name": "tcp-comm",
+                    "protocol": "TCP",
+                    "port": 8786,
+                    "targetPort": "tcp-comm",
+                },
+                {
+                    "name": "http-dashboard",
+                    "protocol": "TCP",
+                    "port": 8787,
+                    "targetPort": "http-dashboard",
+                },
+            ],
+        },
+    }
+
+
 @atexit.register
 def reap_clusters():
     async def _reap_clusters():
