@@ -7,6 +7,7 @@ import kubernetes_asyncio as kubernetes
 
 from uuid import uuid4
 
+from dask.compatibility import entry_points
 from distributed.core import rpc
 
 from dask_kubernetes.common.auth import ClusterAuth
@@ -14,15 +15,27 @@ from dask_kubernetes.common.networking import (
     get_scheduler_address,
 )
 
+# Load operator plugins from other packages
+PLUGINS = []
+for ep in entry_points(group="dask_operator_plugin"):
+    with suppress(AttributeError, ImportError):
+        PLUGINS.append(ep.load())
 
-def build_scheduler_pod_spec(name, spec):
+
+class SchedulerCommError(Exception):
+    """Raised when unable to communicate with a scheduler."""
+
+    pass
+
+
+def build_scheduler_pod_spec(cluster_name, spec):
     return {
         "apiVersion": "v1",
         "kind": "Pod",
         "metadata": {
-            "name": f"{name}-scheduler",
+            "name": f"{cluster_name}-scheduler",
             "labels": {
-                "dask.org/cluster-name": name,
+                "dask.org/cluster-name": cluster_name,
                 "dask.org/component": "scheduler",
                 "sidecar.istio.io/inject": "false",
             },
@@ -31,14 +44,14 @@ def build_scheduler_pod_spec(name, spec):
     }
 
 
-def build_scheduler_service_spec(name, spec):
+def build_scheduler_service_spec(cluster_name, spec):
     return {
         "apiVersion": "v1",
         "kind": "Service",
         "metadata": {
-            "name": f"{name}-service",
+            "name": f"{cluster_name}-scheduler",
             "labels": {
-                "dask.org/cluster-name": name,
+                "dask.org/cluster-name": cluster_name,
             },
         },
         "spec": spec,
@@ -68,7 +81,7 @@ def build_worker_pod_spec(worker_group_name, namespace, cluster_name, uuid, spec
         },
         {
             "name": "DASK_SCHEDULER_ADDRESS",
-            "value": f"tcp://{cluster_name}-service.{namespace}.svc.cluster.local:8786",
+            "value": f"tcp://{cluster_name}-scheduler.{namespace}.svc.cluster.local:8786",
         },
     ]
     for i in range(len(pod_spec["spec"]["containers"])):
@@ -96,7 +109,7 @@ def build_job_pod_spec(job_name, cluster_name, namespace, spec):
     env = [
         {
             "name": "DASK_SCHEDULER_ADDRESS",
-            "value": f"tcp://{cluster_name}-service.{namespace}.svc.cluster.local:8786",
+            "value": f"tcp://{cluster_name}-scheduler.{namespace}.svc.cluster.local:8786",
         },
     ]
     for i in range(len(pod_spec["spec"]["containers"])):
@@ -107,13 +120,19 @@ def build_job_pod_spec(job_name, cluster_name, namespace, spec):
     return pod_spec
 
 
-def build_worker_group_spec(name, spec):
+def build_default_worker_group_spec(cluster_name, spec):
     return {
         "apiVersion": "kubernetes.dask.org/v1",
         "kind": "DaskWorkerGroup",
-        "metadata": {"name": f"{name}-default-worker-group"},
+        "metadata": {
+            "name": f"{cluster_name}-default",
+            "labels": {
+                "dask.org/cluster-name": cluster_name,
+                "dask.org/component": "workergroup",
+            },
+        },
         "spec": {
-            "cluster": name,
+            "cluster": cluster_name,
             "worker": spec,
         },
     }
@@ -123,7 +142,12 @@ def build_cluster_spec(name, worker_spec, scheduler_spec):
     return {
         "apiVersion": "kubernetes.dask.org/v1",
         "kind": "DaskCluster",
-        "metadata": {"name": name},
+        "metadata": {
+            "name": name,
+            "labels": {
+                "dask.org/cluster-name": name,
+            },
+        },
         "spec": {"worker": worker_spec, "scheduler": scheduler_spec},
     }
 
@@ -144,12 +168,22 @@ async def startup(**kwargs):
 
 
 @kopf.on.create("daskcluster")
-async def daskcluster_create(spec, name, namespace, logger, **kwargs):
-    logger.info(
-        f"A DaskCluster has been created called {name} in {namespace} with the following config: {spec}"
-    )
+async def daskcluster_create(name, namespace, logger, patch, **kwargs):
+    """When DaskCluster resource is created set the status.phase.
+
+    This allows us to track that the operator is running.
+    """
+    logger.info(f"DaskCluster {name} created in {namespace}.")
+    patch.status["phase"] = "Created"
+
+
+@kopf.on.field("daskcluster", field="status.phase", new="Created")
+async def daskcluster_create_components(spec, name, namespace, logger, patch, **kwargs):
+    """When the DaskCluster status.phase goes into Pending create the cluster components."""
+    logger.info("Creating Dask cluster components.")
     async with kubernetes.client.api_client.ApiClient() as api_client:
         api = kubernetes.client.CoreV1Api(api_client)
+        custom_api = kubernetes.client.CustomObjectsApi(api_client)
 
         # TODO Check for existing scheduler pod
         scheduler_spec = spec.get("scheduler", {})
@@ -160,10 +194,7 @@ async def daskcluster_create(spec, name, namespace, logger, **kwargs):
             body=data,
         )
         # await wait_for_scheduler(name, namespace)
-        logger.info(
-            f"A scheduler pod has been created called {data['metadata']['name']} in {namespace} \
-            with the following config: {data['spec']}"
-        )
+        logger.info(f"Scheduler pod {data['metadata']['name']} created in {namespace}.")
 
         # TODO Check for existing scheduler service
         data = build_scheduler_service_spec(name, scheduler_spec.get("service"))
@@ -174,26 +205,21 @@ async def daskcluster_create(spec, name, namespace, logger, **kwargs):
         )
         await wait_for_service(api, data["metadata"]["name"], namespace)
         logger.info(
-            f"A scheduler service has been created called {data['metadata']['name']} in {namespace} \
-            with the following config: {data['spec']}"
+            f"Scheduler service {data['metadata']['name']} created in {namespace}."
         )
 
         worker_spec = spec.get("worker", {})
-        data = build_worker_group_spec(name, worker_spec)
-        # TODO: Next line is not needed if we can get worker groups adopted by the cluster
-        kopf.adopt(data)
-        api = kubernetes.client.CustomObjectsApi(api_client)
-        await api.create_namespaced_custom_object(
+        data = build_default_worker_group_spec(name, worker_spec)
+        await custom_api.create_namespaced_custom_object(
             group="kubernetes.dask.org",
             version="v1",
             plural="daskworkergroups",
             namespace=namespace,
             body=data,
         )
-        logger.info(
-            f"A worker group has been created called {data['metadata']['name']} in {namespace} \
-            with the following config: {data['spec']}"
-        )
+        logger.info(f"Worker group {data['metadata']['name']} created in {namespace}.")
+    # TODO Set to "Pending" here and track scheduler readiness before finally setting to "Running"
+    patch.status["phase"] = "Running"
 
 
 @kopf.on.create("daskworkergroup")
@@ -273,6 +299,35 @@ async def retire_workers(
         return [w["metadata"]["name"] for w in workers.items[:-n_workers]]
 
 
+async def get_desired_workers(scheduler_service_name, namespace, logger):
+    # Try gracefully retiring via the HTTP API
+    dashboard_address = await get_scheduler_address(
+        scheduler_service_name,
+        namespace,
+        port_name="http-dashboard",
+    )
+    async with aiohttp.ClientSession() as session:
+        url = f"{dashboard_address}/api/v1/adaptive_target"
+        async with session.get(url) as resp:
+            if resp.status <= 300:
+                desired_workers = await resp.json()
+                return desired_workers["workers"]
+
+    # Otherwise try gracefully retiring via the RPC
+    # Dask version mismatches between the operator and scheduler may cause this to fail in any number of unexpected ways
+    try:
+        comm_address = await get_scheduler_address(
+            scheduler_service_name,
+            namespace,
+        )
+        async with rpc(comm_address) as scheduler_comm:
+            return await scheduler_comm.adaptive_target()
+    except Exception as e:
+        raise SchedulerCommError(
+            "Unable to get number of desired workers from scheduler"
+        ) from e
+
+
 @kopf.on.update("daskworkergroup")
 async def daskworkergroup_update(spec, name, namespace, logger, **kwargs):
     async with kubernetes.client.api_client.ApiClient() as api_client:
@@ -306,7 +361,7 @@ async def daskworkergroup_update(spec, name, namespace, logger, **kwargs):
         if workers_needed < 0:
             worker_ids = await retire_workers(
                 n_workers=-workers_needed,
-                scheduler_service_name=f"{spec['cluster']}-service",
+                scheduler_service_name=f"{spec['cluster']}-scheduler",
                 worker_group_name=name,
                 namespace=namespace,
                 logger=logger,
@@ -329,7 +384,7 @@ async def daskjob_create(spec, name, namespace, logger, **kwargs):
         customobjectsapi = kubernetes.client.CustomObjectsApi(api_client)
         corev1api = kubernetes.client.CoreV1Api(api_client)
 
-        cluster_name = f"{name}-cluster"
+        cluster_name = f"{name}"
         cluster_spec = build_cluster_spec(
             cluster_name,
             spec["cluster"]["spec"]["worker"],
@@ -344,7 +399,7 @@ async def daskjob_create(spec, name, namespace, logger, **kwargs):
             body=cluster_spec,
         )
         logger.info(
-            f"A cluster for the job {name} has been created called {cluster_spec['metadata']['name']} in {namespace}"
+            f"Cluster {cluster_spec['metadata']['name']} for job {name} created in {namespace}."
         )
 
         job_pod_spec = build_job_pod_spec(
@@ -360,17 +415,82 @@ async def daskjob_create(spec, name, namespace, logger, **kwargs):
         )
 
 
-@kopf.on.field("pod", field="status.phase", labels={"dask.org/component": "job-runner"})
+@kopf.on.field(
+    "pod",
+    field="status.phase",
+    labels={"dask.org/component": "job-runner"},
+    new="Succeeded",
+)
 async def handle_runner_status_change(meta, new, namespace, logger, **kwargs):
-    logger.info(f"Job now in phase {new}.")
-    if new == "Succeeded":
-        logger.info("Job succeeded, deleting Dask cluster.")
-        async with kubernetes.client.api_client.ApiClient() as api_client:
-            customobjectsapi = kubernetes.client.CustomObjectsApi(api_client)
-            await customobjectsapi.delete_namespaced_custom_object(
-                group="kubernetes.dask.org",
-                version="v1",
-                plural="daskclusters",
-                namespace=namespace,
-                name=meta["labels"]["dask.org/cluster-name"],
-            )
+    logger.info("Job succeeded, deleting Dask cluster.")
+    async with kubernetes.client.api_client.ApiClient() as api_client:
+        customobjectsapi = kubernetes.client.CustomObjectsApi(api_client)
+        await customobjectsapi.delete_namespaced_custom_object(
+            group="kubernetes.dask.org",
+            version="v1",
+            plural="daskclusters",
+            namespace=namespace,
+            name=meta["labels"]["dask.org/cluster-name"],
+        )
+
+
+@kopf.on.create("daskautoscaler")
+async def daskautoscaler_create(spec, name, namespace, logger, **kwargs):
+    """When an autoscaler is created make it a child of the associated cluster for cascade deletion."""
+    async with kubernetes.client.api_client.ApiClient() as api_client:
+        api = kubernetes.client.CustomObjectsApi(api_client)
+        cluster = await api.get_namespaced_custom_object(
+            group="kubernetes.dask.org",
+            version="v1",
+            plural="daskclusters",
+            namespace=namespace,
+            name=spec["cluster"],
+        )
+        new_spec = dict(spec)
+        kopf.adopt(new_spec, owner=cluster)
+        api.api_client.set_default_header(
+            "content-type", "application/merge-patch+json"
+        )
+        await api.patch_namespaced_custom_object(
+            group="kubernetes.dask.org",
+            version="v1",
+            plural="daskautoscalers",
+            namespace=namespace,
+            name=name,
+            body=new_spec,
+        )
+        logger.info(f"Successfully adopted by {spec['cluster']}")
+
+
+@kopf.timer("daskautoscaler", interval=5.0)
+async def daskautoscaler_adapt(spec, name, namespace, logger, **kwargs):
+    # Ask the scheduler for the desired number of worker
+    try:
+        desired_workers = await get_desired_workers(
+            scheduler_service_name=f"{spec['cluster']}-scheduler",
+            namespace=namespace,
+            logger=logger,
+        )
+    except SchedulerCommError:
+        logger.error("Unable to get desired number of workers from scheduler.")
+        return
+
+    # Ensure the desired number is within the min and max
+    desired_workers = max(spec["minimum"], desired_workers)
+    desired_workers = min(spec["maximum"], desired_workers)
+
+    # Update the default DaskWorkerGroup
+    # TODO Only update if the value has changed
+    async with kubernetes.client.api_client.ApiClient() as api_client:
+        customobjectsapi = kubernetes.client.CustomObjectsApi(api_client)
+        customobjectsapi.api_client.set_default_header(
+            "content-type", "application/merge-patch+json"
+        )
+        await customobjectsapi.patch_namespaced_custom_object_scale(
+            group="kubernetes.dask.org",
+            version="v1",
+            plural="daskworkergroups",
+            namespace=namespace,
+            name=f"{spec['cluster']}-default",
+            body={"spec": {"replicas": desired_workers}},
+        )
