@@ -18,7 +18,7 @@ def gen_cluster(k8s_cluster):
     @asynccontextmanager
     async def cm():
         cluster_path = os.path.join(DIR, "resources", "simplecluster.yaml")
-        cluster_name = "simple-cluster"
+        cluster_name = "simple"
 
         # Create cluster resource
         k8s_cluster.kubectl("apply", "-f", cluster_path)
@@ -75,13 +75,22 @@ def test_operator_runs(kopf_runner):
     assert runner.exception is None
 
 
+def test_operator_plugins(kopf_runner):
+    with kopf_runner as runner:
+        pass
+
+    assert runner.exit_code == 0
+    assert runner.exception is None
+    assert "Plugin 'noop' running." in runner.stdout
+
+
 @pytest.mark.asyncio
 async def test_scalesimplecluster(k8s_cluster, kopf_runner, gen_cluster):
     with kopf_runner as runner:
         async with gen_cluster() as cluster_name:
-            scheduler_pod_name = "simple-cluster-scheduler"
-            worker_pod_name = "simple-cluster-default-worker-group-worker"
-            service_name = "simple-cluster-service"
+            scheduler_pod_name = "simple-scheduler"
+            worker_pod_name = "simple-default-worker"
+            service_name = "simple-scheduler"
             while scheduler_pod_name not in k8s_cluster.kubectl("get", "pods"):
                 await asyncio.sleep(0.1)
             while service_name not in k8s_cluster.kubectl("get", "svc"):
@@ -100,14 +109,14 @@ async def test_scalesimplecluster(k8s_cluster, kopf_runner, gen_cluster):
                         "scale",
                         "--replicas=5",
                         "daskworkergroup",
-                        "simple-cluster-default-worker-group",
+                        "simple-default",
                     )
                     await client.wait_for_workers(5)
                     k8s_cluster.kubectl(
                         "scale",
                         "--replicas=3",
                         "daskworkergroup",
-                        "simple-cluster-default-worker-group",
+                        "simple-default",
                     )
                     await client.wait_for_workers(3)
 
@@ -117,9 +126,9 @@ async def test_scalesimplecluster(k8s_cluster, kopf_runner, gen_cluster):
 async def test_simplecluster(k8s_cluster, kopf_runner, gen_cluster):
     with kopf_runner as runner:
         async with gen_cluster() as cluster_name:
-            scheduler_pod_name = "simple-cluster-scheduler"
-            worker_pod_name = "simple-cluster-default-worker-group-worker"
-            service_name = "simple-cluster-service"
+            scheduler_pod_name = "simple-scheduler"
+            worker_pod_name = "simple-default-worker"
+            service_name = "simple-scheduler"
 
             while scheduler_pod_name not in k8s_cluster.kubectl("get", "pods"):
                 await asyncio.sleep(0.1)
@@ -161,10 +170,6 @@ async def test_simplecluster(k8s_cluster, kopf_runner, gen_cluster):
             assert "WORKER_ENV" in worker_env
             assert cluster_name
 
-    assert "A DaskCluster has been created" in runner.stdout
-    assert "A scheduler pod has been created" in runner.stdout
-    assert "A worker group has been created" in runner.stdout
-
 
 @pytest.mark.asyncio
 async def test_job(k8s_cluster, kopf_runner, gen_job):
@@ -172,15 +177,14 @@ async def test_job(k8s_cluster, kopf_runner, gen_job):
         async with gen_job() as job:
             assert job
 
-            cluster_name = f"{job}-cluster"
             runner_name = f"{job}-runner"
 
             # Assert that cluster is created
-            while cluster_name not in k8s_cluster.kubectl("get", "daskclusters"):
+            while job not in k8s_cluster.kubectl("get", "daskclusters"):
                 await asyncio.sleep(0.1)
 
             # Assert job pod is created
-            while runner_name not in k8s_cluster.kubectl("get", "po"):
+            while job not in k8s_cluster.kubectl("get", "po"):
                 await asyncio.sleep(0.1)
 
             # Assert job pod runs to completion (will fail if doesn't connect to cluster)
@@ -188,7 +192,7 @@ async def test_job(k8s_cluster, kopf_runner, gen_job):
                 await asyncio.sleep(0.1)
 
             # Assert cluster is removed on completion
-            while cluster_name in k8s_cluster.kubectl("get", "daskclusters"):
+            while job in k8s_cluster.kubectl("get", "daskclusters"):
                 await asyncio.sleep(0.1)
 
     assert "A DaskJob has been created" in runner.stdout
